==> spring-boot-starter-model-0.0.5-20210511.123047-10-sources.jar: cn/com/duiba/spring/boot/starter/autoconfigure/tensorflow/service/impl/AlgoTFModelProxyImpl.java <==
package cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.service.impl;

import cn.com.duiba.nezha.alg.model.tf.LocalTFModel;
import cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.enums.AlgoTFModelStatusEnum;
import cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.model.AlgoTFModel;
import cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.support.ChooseTFModelStrategy;
import cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.support.StrategyFactory;
import cn.com.duiba.spring.boot.starter.autoconfigure.service.AlgoTFModelProxy;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import java.util.Random;

@Slf4j
@Data
public class AlgoTFModelProxyImpl implements AlgoTFModelProxy {

    private static Random random = new Random();

    // 模型key
    private String tfKey;
    // 模型1
    private AlgoTFModel algoTFModel1;
    // 模型2
    private AlgoTFModel algoTFModel2;
    // 模型更新时间
    private long tfModelUpdateTime;

    AlgoTFModelProxyImpl(String tfKey) throws Exception {
        this.tfKey = tfKey;
        LocalTFModel localTFModel = new LocalTFModel();
        try {
            localTFModel.loadModel(tfKey);
        } catch (Exception e) {
            log.warn("localTFModel loadModel error...", e);
            throw e;
        }
        // 初始化投入生产的模型
        algoTFModel1 = new AlgoTFModel(localTFModel, AlgoTFModelStatusEnum.RUNNING.getCode(), tfKey + "-A");
        // 初始化准备投入生产的模型
        algoTFModel2 = new AlgoTFModel(AlgoTFModelStatusEnum.STOP.getCode(), tfKey + "-B");
    }

    @Override
    public LocalTFModel chooseTFModel() {
        // case1
        if (!algoTFModel1.isRunning() && !algoTFModel2.isRunning()) {
            return null;
        }

        // case2
        if (algoTFModel1.isRunning() && !algoTFModel2.isRunning()) {
            return algoTFModel1.getLocalTFModel();
        }

        // case3
        if (!algoTFModel1.isRunning() && algoTFModel2.isRunning()) {
            return algoTFModel2.getLocalTFModel();
        }

        AlgoTFModel newTFModel;
        AlgoTFModel oldTFModel;
        if (algoTFModel1.getLoadTime().compareTo(algoTFModel2.getLoadTime()) > 0) {
            newTFModel = algoTFModel1;
            oldTFModel = algoTFModel2;
        } else {
            newTFModel = algoTFModel2;
            oldTFModel = algoTFModel1;
        }
        Long loadTime = newTFModel.getLoadTime();


//        if (true) {
//            return newTFModel.getLocalTFModel();
//        }
//        return oldTFModel.getLocalTFModel();

        if (System.currentTimeMillis() - loadTime >= 90000) {
            return newTFModel.getLocalTFModel();
        }

        ChooseTFModelStrategy strategy = StrategyFactory.getStrategy(tfKey);
        if (strategy.tryAcquire(loadTime)) {
            return newTFModel.getLocalTFModel();
        }
        return oldTFModel.getLocalTFModel();



//        if (System.currentTimeMillis() - loadTime < 90000) {
//            // 获取分流比例
//            double weight = getWeight(loadTime);
//            // 获取tf模型
//            return getTFModelWithWeight(newTFModel, oldTFModel, weight);
//        }
//
//        return newTFModel.getLocalTFModel();
    }

    @Override
    public void closeTFModel() {

        // 如果模型1超过5分钟没有被访问，进行关闭
        if (algoTFModel1.isAllowClose()) {
            algoTFModel1.doCloseTFModel();
        }

        // 如果模型2超过5分钟没有被访问，进行关闭
        if (algoTFModel2.isAllowClose()) {
            algoTFModel2.doCloseTFModel();
        }

    }

    @Override
    public boolean updateTFModel() {
        // case1：模型1停止运行，并且模型2停止运行
        // 则代表此key对应的模型已经停止使用了, 不做处理，继续下一个循环逻辑
        if (!algoTFModel1.isRunning() && !algoTFModel2.isRunning()) {
            return false;
        }
        // case2：模型1正在运行，模型2停止运行
        // 则立刻加载新模型到模型2上
        LocalTFModel localTFModel = new LocalTFModel();
        if (algoTFModel1.isRunning() && !algoTFModel2.isRunning() && algoTFModel1.isLatestVersion(localTFModel, tfKey)) {
            algoTFModel2.loadTFModel(localTFModel, tfKey);
            tfModelUpdateTime = System.currentTimeMillis();
            return true;
        }

        // case3：模型1停止运行，模型2正在运行
        // 则立刻加载新模型到模型1上
        if (!algoTFModel1.isRunning() && algoTFModel2.isRunning() && algoTFModel2.isLatestVersion(localTFModel, tfKey)) {
            algoTFModel1.loadTFModel(localTFModel, tfKey);
            tfModelUpdateTime = System.currentTimeMillis();
            return true;
        }
        return false;
    }

    @Override
    public boolean hasTwoRunningModel() {
        return algoTFModel1.isRunning() && algoTFModel2.isRunning();
    }

    /**
     * 获取分流比例
     * @param loadTime
     * @return
     */
    private static double getWeight(Long loadTime){
        long deltaTime = System.currentTimeMillis() - loadTime;
        return (double)(deltaTime / 10000);
    }

    /**
     * 分流获取模型
     * @param newTFModel
     * @param oldTFModel
     * @param weight
     * @return
     */
    private LocalTFModel getTFModelWithWeight(AlgoTFModel newTFModel, AlgoTFModel oldTFModel, double weight) {
        boolean shunt = random.nextDouble() * 100 <= Math.sqrt(weight);
        if (shunt) {
            return newTFModel.getLocalTFModel();
        }
        return oldTFModel.getLocalTFModel();
    }

}

==> spring-boot-starter-model-0.0.6-20210512.082607-25-sources.jar: cn/com/duiba/spring/boot/starter/autoconfigure/tensorflow/service/impl/AlgoTFModelProxyImpl.java <==
package cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.service.impl;

import cn.com.duiba.nezha.alg.model.tf.LocalTFModel;
import cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.enums.AlgoTFModelStatusEnum;
import cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.model.AlgoTFModel;
import cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.support.ChooseTFModelStrategy;
import cn.com.duiba.spring.boot.starter.autoconfigure.tensorflow.support.StrategyFactory;
import cn.com.duiba.spring.boot.starter.autoconfigure.service.AlgoTFModelProxy;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import java.util.Random;

@Slf4j
@Data
public class AlgoTFModelProxyImpl implements AlgoTFModelProxy {

    private static Random random = new Random();

    // 模型key
    private String tfKey;
    // 模型1
    private AlgoTFModel algoTFModel1;
    // 模型2
    private AlgoTFModel algoTFModel2;
    // 模型更新时间
    private long tfModelUpdateTime;

    AlgoTFModelProxyImpl(String tfKey) throws Exception {
        this.tfKey = tfKey;
        LocalTFModel localTFModel = new LocalTFModel();
        try {
            localTFModel.loadModel(tfKey);
        } catch (Exception e) {
            log.warn("localTFModel loadModel error...", e);
            throw e;
        }
        // 初始化投入生产的模型
        algoTFModel1 = new AlgoTFModel(localTFModel, AlgoTFModelStatusEnum.RUNNING.getCode(), tfKey + "-A");
        // 初始化准备投入生产的模型
        algoTFModel2 = new AlgoTFModel(AlgoTFModelStatusEnum.STOP.getCode(), tfKey + "-B");
    }

    @Override
    public LocalTFModel chooseTFModel() {
        // case1
        if (!algoTFModel1.isRunning() && !algoTFModel2.isRunning()) {
            return null;
        }

        // case2
        if (algoTFModel1.isRunning() && !algoTFModel2.isRunning()) {
            return algoTFModel1.getLocalTFModel();
        }

        // case3
        if (!algoTFModel1.isRunning() && algoTFModel2.isRunning()) {
            return algoTFModel2.getLocalTFModel();
        }

        AlgoTFModel newTFModel;
        AlgoTFModel oldTFModel;
        if (algoTFModel1.getLoadTime().compareTo(algoTFModel2.getLoadTime()) > 0) {
            newTFModel = algoTFModel1;
            oldTFModel = algoTFModel2;
        } else {
            newTFModel = algoTFModel2;
            oldTFModel = algoTFModel1;
        }
        Long loadTime = newTFModel.getLoadTime();


//        if (true) {
//            return newTFModel.getLocalTFModel();
//        }
//        return oldTFModel.getLocalTFModel();

        if (System.currentTimeMillis() - loadTime >= 90000) {
            return newTFModel.getLocalTFModel();
        }

        ChooseTFModelStrategy strategy = StrategyFactory.getStrategy();
        if (strategy.tryAcquire(loadTime)) {
            return newTFModel.getLocalTFModel();
        }
        return oldTFModel.getLocalTFModel();



//        if (System.currentTimeMillis() - loadTime < 90000) {
//            // 获取分流比例
//            double weight = getWeight(loadTime);
//            // 获取tf模型
//            return getTFModelWithWeight(newTFModel, oldTFModel, weight);
//        }
//
//        return newTFModel.getLocalTFModel();
    }

    @Override
    public void closeTFModel() {

        // 如果模型1超过5分钟没有被访问，进行关闭
        if (algoTFModel1.isAllowClose()) {
            algoTFModel1.doCloseTFModel();
        }

        // 如果模型2超过5分钟没有被访问，进行关闭
        if (algoTFModel2.isAllowClose()) {
            algoTFModel2.doCloseTFModel();
        }

    }

    @Override
    public boolean updateTFModel() {
        // case1：模型1停止运行，并且模型2停止运行
        // 则代表此key对应的模型已经停止使用了, 不做处理，继续下一个循环逻辑
        if (!algoTFModel1.isRunning() && !algoTFModel2.isRunning()) {
            return false;
        }
        // case2：模型1正在运行，模型2停止运行
        // 则立刻加载新模型到模型2上
        LocalTFModel localTFModel = new LocalTFModel();
        if (algoTFModel1.isRunning() && !algoTFModel2.isRunning() && algoTFModel1.isLatestVersion(localTFModel, tfKey)) {
            algoTFModel2.loadTFModel(localTFModel, tfKey);
            tfModelUpdateTime = System.currentTimeMillis();
            return true;
        }

        // case3：模型1停止运行，模型2正在运行
        // 则立刻加载新模型到模型1上
        if (!algoTFModel1.isRunning() && algoTFModel2.isRunning() && algoTFModel2.isLatestVersion(localTFModel, tfKey)) {
            algoTFModel1.loadTFModel(localTFModel, tfKey);
            tfModelUpdateTime = System.currentTimeMillis();
            return true;
        }
        return false;
    }

    @Override
    public boolean hasTwoRunningModel() {
        return algoTFModel1.isRunning() && algoTFModel2.isRunning();
    }

    /**
     * 获取分流比例
     * @param loadTime
     * @return
     */
    private static double getWeight(Long loadTime){
        long deltaTime = System.currentTimeMillis() - loadTime;
        return (double)(deltaTime / 10000);
    }

    /**
     * 分流获取模型
     * @param newTFModel
     * @param oldTFModel
     * @param weight
     * @return
     */
    private LocalTFModel getTFModelWithWeight(AlgoTFModel newTFModel, AlgoTFModel oldTFModel, double weight) {
        boolean shunt = random.nextDouble() * 100 <= Math.sqrt(weight);
        if (shunt) {
            return newTFModel.getLocalTFModel();
        }
        return oldTFModel.getLocalTFModel();
    }

}
